"""ftp消息通道"""

import io
import pathlib
from typing import Tuple
from queue import Empty
from vxutils import logger, vxFTPConnector, vxtime
from vxsched.core import vxEventBroker
from vxsched.event import vxEvent


def _to_event_rank(filename: str) -> tuple:
    """转换为 (timestamp, filename)的组合"""
    return float(pathlib.Path(filename).name.split("||")[0]), filename


class vxFTPEventBroker(vxEventBroker):
    """ftp消息分发"""

    _connections = {}
    channels = {}

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 21,
        user: str = "",
        passwd: str = "",
        root_dir: str = "/",
    ):
        super().__init__()
        key = hash(f"ftp://{user}:{passwd}@{host}:{port}")
        if key not in self.__class__._connections:
            self.__class__._connections[key] = vxFTPConnector(host, port, user, passwd)
            logger.info(f"创建FTP连接：{self.__class__._connections[key]}")

        self._ftp_conn = self.__class__._connections[key]

        self._root_dir = pathlib.PurePath(root_dir).as_posix()
        remote_dirs = self._ftp_conn.list(self._root_dir)
        for channel_name in remote_dirs:
            self.__class__.channels[channel_name] = pathlib.Path(
                self._root_dir, channel_name
            ).as_posix()

    def _get_lastest_event_filename(self, channel_name: str) -> Tuple[float, str]:
        """获取最近的事件文件名称

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
            Tuple[float, str] -- 返回 下一次运行时间以及事件文件路径。
        """
        remote_files = self._ftp_conn.list(
            pathlib.Path(self._root_dir, channel_name).as_posix()
        )

        if not remote_files:
            raise Empty

        return min(map(_to_event_rank, remote_files))

    def get(self, channel_name: str) -> vxEvent:
        """获取一个消息

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一个vxEvent消息

        Exceptions:
           Empty : 队列为空，抛出Empty异常

        """
        try:
            _, next_event_file = self._get_lastest_event_filename(channel_name)
        except Empty as err:
            vxtime.sleep(1)
            raise Empty from err

        with io.BytesIO() as bfp:
            is_download = self._ftp_conn.download(next_event_file, bfp)
            if is_download is False:
                raise ConnectionError(f"下载{next_event_file}发生错误..")
            event = vxEvent.unpack(bfp.getvalue())
            self._ftp_conn.delete(next_event_file)
            return event

    def put(self, channel_name: str, event: vxEvent) -> None:
        """将event消息放入消息队列中

        Arguments:
            channel_name {str} -- 消息通道名称
            event {vxEvent} -- 消息

        Returns:
            None
        """

        try:
            remote_dir = self.__class__.channels.get(channel_name, None)
            if remote_dir is None:
                remote_dir = pathlib.Path(self._root_dir, channel_name).as_posix()
                self._ftp_conn.mkdir(remote_dir)
                self.__class__.channels[channel_name] = remote_dir

            picked_event = vxEvent.pack(event)
            remote_file = pathlib.Path(
                remote_dir, f"{event.next_trigger_dt}||{event.id}.pkl"
            )

            with io.BytesIO(picked_event) as bfp:
                self._ftp_conn.upload(bfp, remote_file)
                logger.debug(
                    f"{self} put event.type({event.type}) to remote({remote_file}). {event}"
                )

        except Exception as err:
            logger.error(
                f"{self} put event.type({event.type}) to remote({remote_file}). error: {err}",
                exc_info=True,
            )

    def get_first_event_trigger_dt(self, channel_name: str) -> float:
        """获取最近一次执行时间

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一次执行时间

        Exceptions:
           Empty : 队列为空

        """
        first_event_trigger_dt, _ = self._get_lastest_event_filename(channel_name)
        return first_event_trigger_dt

    def clear(self, channel_name: str) -> None:
        """清空队列

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
            None
        """
        remote_dir = self.__class__.channels.get(channel_name)
        if remote_dir is None:
            return

        remote_files = self._ftp_conn.list(remote_dir)
        for remote_file in remote_files:
            self._ftp_conn.delete(pathlib.Path(remote_dir, remote_file).as_posix())
