"""ZMQ 消息通道"""


from collections import defaultdict
from queue import Empty, PriorityQueue
from typing import Optional
import zmq
from vxutils import logger, ZMQClient, create_server, vxtime
from vxsched.event import vxEvent
from vxsched.core import vxEventBroker

_ZMQCHANNEL = "__zmqchannel__"


class vxZMQEventBroker(vxEventBroker):
    """ZMQ消息通道"""

    def __init__(
        self,
        url: str = "tcp://127.0.0.1:5555",
        server_key: Optional[str] = None,
    ):
        super(vxZMQEventBroker, self).__init__()
        self._url = url
        self._client = ZMQClient(url, server_key)

    def __str__(self):
        return f"{self.__class__.__name__}-id({id(self)} with url: {self._url}"

    __repr__ = __str__

    def get(self, channel_name: str) -> vxEvent:
        """获取一个消息

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一个vxEvent消息

        Exceptions:
           Empty : 队列为空，抛出Empty异常

        """
        try:
            req_event = vxEvent(
                type=f"{_ZMQCHANNEL}.__get__", data="", channel=channel_name
            )
            event = self._client.send(req_event)
        except zmq.ZMQError as e:
            logger.error(f"发送消息错误：{e}", exc_info=True)
            event = None

        if not event:
            vxtime.sleep(0.1)
            raise Empty
        return event

    def put(self, channel_name: str, event: vxEvent) -> None:
        """将event消息放入消息队列中

        Arguments:
            channel_name {str} -- 消息通道名称
            event {vxEvent} -- 消息

        Returns:
            None
        """

        try:
            req_event = vxEvent(
                type=f"{_ZMQCHANNEL}.__put__",
                channel=channel_name,
                data=event,
            )
            self._client.send(req_event)
        except zmq.ZMQError as e:
            logger.error(f"发送消息错误：{e}", exc_info=True)

    def get_first_event_trigger_dt(self, channel_name: str) -> float:
        """获取最近一次执行时间

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一次执行时间

        Exceptions:
           Empty : 队列为空

        """
        try:
            req_event = vxEvent(
                type=f"{_ZMQCHANNEL}.__triggerdt__",
                channel=channel_name,
                data="",
            )
            first_event_trigger_dt = self._client.send(req_event)
        except zmq.ZMQError as e:
            logger.error(f"发送消息错误：{e}", exc_info=True)
            first_event_trigger_dt = None
        if not first_event_trigger_dt:
            raise Empty
        return first_event_trigger_dt


class vxZMQEventBrokerServer:
    """消息服务总线

    vxhandlers : 事件处理器
    url : 事件总线地址
    secret_key : 加密密钥
    """

    def __init__(
        self, url: str = "tcp://*:5555", secret_key: Optional[str] = None
    ) -> None:
        self._url: str = url
        self._secret_key: str = secret_key
        self._queues: dict = defaultdict(PriorityQueue)
        self._active: bool = False

    def _run(self) -> None:

        with create_server(self._url, zmq.REP, self._secret_key) as server:
            while self._active:
                if not server.poll(100):
                    continue

                event = server.recv_pyobj()
                ret = ""
                try:
                    if not isinstance(event, vxEvent):
                        logger.warning(f"收到一个非vxEvent对象: {event}")
                        continue

                    if event.type == f"{_ZMQCHANNEL}.__put__":
                        recive_event = event.data
                        self._queues[recive_event.channel].put_nowait(recive_event)

                    elif event.type == f"{_ZMQCHANNEL}.__get__":
                        ret = self._queues[event.channel].get_nowait()
                    elif event.type == f"{_ZMQCHANNEL}.__triggerdt__":
                        if self._queues[event.channel].empty():
                            raise Empty
                        ret = self._queues[event.channel].queue[0].next_trigger_dt
                    else:
                        raise ValueError(f"未知事件,无法处理,{event}")
                except Empty:
                    ret = ""
                except Exception as e:
                    logger.error(f"发生运行错误: {e}", exc_info=True)
                    ret = e
                finally:
                    server.send_pyobj(ret)

    def start(self) -> None:
        """启动zmqeventbroker server服务"""
        if self._active:
            logger.warning(f"{self._url} 已经激活，不能再次激活.")
            return

        self._active = True
        try:
            self._run()
        finally:
            self.stop()

    def stop(self):
        """暂停zmqeventbroker server服务"""
        self._active = False
