# encoding=utf-8
""" 策略handler模块 """


import uuid
import abc

from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor
import os
import time
import pathlib
import importlib
import contextlib

from functools import wraps
from collections import defaultdict
from queue import Empty, PriorityQueue
from abc import ABCMeta
from typing import Any, Callable, Optional, Union, Dict
from vxutils import logger, vxtime, vxToolBox
from vxsched.event import vxEvent, TriggerStatus, vxTrigger
from vxsched.context import vxContext


__all__ = [
    "vxEventBroker",
    "vxLocalEventBroker",
    "vxChannel",
    "vxSched",
    "init_scheduler",
    "vxscheduler",
]


class vxEventBroker(metaclass=ABCMeta):
    """消息队列"""

    queue_pools: Dict[str, PriorityQueue] = defaultdict(PriorityQueue)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}-(id-{id(self)})"

    __repr__ = __str__

    @abc.abstractmethod
    def get(self, channel_name: str) -> vxEvent:
        """获取一个消息

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一个vxEvent消息

        Exceptions:
           Empty : 队列为空，抛出Empty异常

        """

    @abc.abstractmethod
    def put(self, channel_name: str, event: vxEvent) -> None:
        """将event消息放入消息队列中

        Arguments:
            channel_name {str} -- 消息通道名称
            event {vxEvent} -- 消息

        Returns:
            None
        """

    @abc.abstractmethod
    def get_first_event_trigger_dt(self, channel_name: str) -> float:
        """获取最近一次执行时间

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
           返回最近一次执行时间

        Exceptions:
           Empty : 队列为空

        """

    def clear(self, channel_name: str) -> None:
        """清空队列

        Arguments:
            channel_name {str} -- 消息通道名称

        Returns:
            None
        """


class vxLocalEventBroker(vxEventBroker):
    """本地消息队列"""

    def get(self, channel_name: str) -> vxEvent:
        queue = self.__class__.queue_pools[channel_name]
        return queue.get(timeout=0.1)

    def put(self, channel_name: str, event: vxEvent) -> None:
        queue = self.__class__.queue_pools[channel_name]
        return queue.put_nowait(event)

    def get_first_event_trigger_dt(self, channel_name: str) -> float:
        queue = self.__class__.queue_pools[channel_name]
        if queue.empty():
            raise Empty
        return queue.queue[0].next_trigger_dt

    def clear(self, channel_name: str) -> None:
        queue = self.__class__.queue_pools[channel_name]
        with contextlib.suppress(Empty):
            while True:
                queue.get_nowait()


class vxChannel:
    """消息通道"""

    def __init__(
        self, channel_name: str, event_broker: Optional[vxEventBroker] = None
    ) -> None:
        self._channel_name = channel_name
        self._event_broker = event_broker or vxLocalEventBroker()

    def __str__(self):
        return f"< vxChannel ({self.name}) with queue: ({self._event_broker}) >"

    __repr__ = __str__

    def __eq__(self, other: "vxChannel") -> bool:
        return self.name == other.name and self._event_broker == other._event_broker

    @property
    def name(self) -> str:
        """通道名称"""
        return self._channel_name

    def put(
        self,
        event: Union[str, vxEvent],
        data: Any = None,
        trigger: Optional[vxTrigger] = None,
        priority=10,
        **kwargs,
    ) -> None:
        """发送消息

        Arguments:
            event {Union[str, vxEvent]} -- 要推送消息或消息类型

        Keyword Arguments:
            data {Any} -- 消息数据信息 (default: {None})
            trigger {Optional[vxTrigger]} -- 消息触发器 (default: {None})
            priority {int} -- 优先级，越小优先级越高 (default: {10})
        """

        if isinstance(event, str):
            send_event = vxEvent(
                type=event,
                data=data,
                trigger=trigger,
                priority=priority,
                **kwargs,
            )

        else:
            send_event = event
        send_event.channel = self.name

        if send_event.trigger:
            if send_event.trigger.status == TriggerStatus.Completed:
                return

            # 每放一次，都会跳一次
            next(
                send_event.trigger,
            )
            send_event.next_trigger_dt = send_event.trigger.last_trigger_dt
        logger.debug(f"发送event: {send_event}")
        self._event_broker.put(self.name, send_event)

    def get(self, timestamp_: Optional[float] = None) -> Optional[vxEvent]:
        """获取消息

        Keyword Arguments:
            timestamp_ {Optional[float]} -- 时间戳，获取小于此时间戳的最优先消息 (default: {None})

        Returns:
            Optional[vxEvent] -- 返回消息或空
        """
        if timestamp_ is None:
            timestamp_ = vxtime.now()

        first_event_trigger_dt = self._event_broker.get_first_event_trigger_dt(
            channel_name=self.name
        )

        if first_event_trigger_dt >= timestamp_:
            raise Empty

        event = self._event_broker.get(self.name)

        if event.trigger and event.trigger.status != TriggerStatus.Completed:
            event.next_trigger_dt = next(event.trigger)
            self._event_broker.put(self.name, event)

        return event

    def clear(self):
        """清空队列所有消息"""
        self._event_broker.clear(self.name)


class vxSched:
    """调度器"""

    def __init__(self, context: Optional[vxContext] = None) -> None:

        # 监听brokers
        self._bind_brokers = []

        # 执行handlers
        self._handlers = {}
        # 上下文
        self._context = context or vxContext()

        self._active = False

    @property
    def context(self):
        """获取上下文"""
        return self._context

    def register(self, event_type: str, handler: Callable) -> None:
        """注册事件处理函数

        Arguments:
            event_type {str} -- 消息类型
            handler {Callable} -- 消息处理函数,handler(context, event, tools=None)

        Raises:
            ValueError: handler类型错误
        """

        if not callable(handler):
            raise ValueError(f"{handler.__name__} is not callable.")

        handlers = self._handlers.pop(event_type, [])
        handler_names = [handler.__name__ for handler in handlers]

        if (handler.__name__ not in handler_names) and (handler not in handlers):
            handlers.append(handler)
            logger.info(
                f"vxScheduler register event type: {event_type}, handler: {handler}"
            )
        self._handlers[event_type] = handlers

    def unregister(self, event_type: str, handler: Optional[Callable] = None) -> None:
        """取消注册事件处理函数

        Arguments:
            event_type {str} -- 消息类型

        Keyword Arguments:
            handler {Optional[Callable]} -- 消息事件处理函数 (default: {None})
        """

        logger.info(
            f"vxScheduler unregister event type: {event_type}, handler: {handler}"
        )
        handlers = self._handlers.pop(event_type, [])
        if handler:
            if handler in handlers:
                handlers.remove(handler)

            if handlers:
                self._handlers[event_type] = handlers

    def unregister_handler(self, handler: Callable) -> None:
        """取消监听特定函数

        Arguments:
            handler {Callable} -- 待取消监听的函数
        """

        handlers = self._handlers

        for event_type, handlers in handlers.items():
            if handler in handlers:
                handlers.remove(handler)
                if handlers:
                    self._handlers[event_type] = handlers
                else:
                    self._handlers.pop(event_type)

                logger.info(
                    f"vxScheduler unregister event type: {event_type}, handler: {handler}"
                )

    def trigger(self, event: vxEvent) -> Any:
        """触发一个消息

        Arguments:
            event {vxEvent} -- 待触发的消息类型

        Returns:
            Any -- 所有handler
        """
        handlers = self._handlers.get(str(event.type), [])
        return list(map(lambda hdl: hdl(self.context, event), handlers))

    def __call__(self, event_type, time_limit=1):
        def deco(func):
            @wraps(func)
            def _event_handler(context, event):
                try:
                    start = time.perf_counter()
                    ret = func(context, event)
                    cost_time = time.perf_counter() - start
                    if cost_time > time_limit:
                        logger.warning(
                            f"vxScheduler:{func.__name__} 耗时 {cost_time*1000:,.2f}ms 超出预定时间 {time_limit*1000:,.2f}ms. event={event}"
                        )

                except Exception as err:
                    logger.error(
                        f"vxScheduler:{func.__name__} error: {err},event={event}",
                        exc_info=True,
                    )
                    ret = None

                return ret

            self.register(event_type, _event_handler)
            return func

        return deco

    def load_modules(self, mod_path):
        """加载策略目录"""
        if not os.path.exists(mod_path):
            logger.warning(msg=f"{mod_path} is not exists")
            return

        modules = os.listdir(mod_path)
        logger.info(f"loading strategy dir: {mod_path}.")
        logger.info("=" * 80)
        for mod in modules:
            if (not mod.startswith("__")) and mod.endswith(".py"):
                try:
                    loader = importlib.machinery.SourceFileLoader(
                        mod, os.path.join(mod_path, mod)
                    )
                    spec = importlib.util.spec_from_loader(loader.name, loader)
                    strategy_mod = importlib.util.module_from_spec(spec)
                    loader.exec_module(strategy_mod)
                    logger.info(f"Load Module: {strategy_mod} Sucess.")
                    logger.info("+" * 80)
                except Exception as err:
                    logger.error(f"Load Module: {mod} Failed. {err}", exc_info=True)
                    logger.error("-" * 80)

    def bind_eventbroker(self, eventbroker: vxEventBroker) -> None:
        """绑定event broker

        Arguments:
            eventbroker {vxEventBroker} -- 待绑定的event broker
        """
        self._bind_brokers.append(eventbroker)

    def _run(self, *channels) -> None:
        """消息引擎运行函数"""
        if isinstance(channels[0], list):
            channels = channels[0]

        while self._active:
            for channel in channels:
                with contextlib.suppress(Empty):
                    event = channel.get()
                    self.trigger(event)
            vxtime.sleep(0)

    def start(
        self,
        channel_name: str = "local",
        workers: int = 1,
        executor: Executor = None,
    ) -> None:
        """开始运行vxScheduler

        Keyword Arguments:
            channel_name {str} -- 消息通道名称 (default: {"local"})
            workers {int} --  线程数量 (default: {1})
            executor {Executor} -- 执行器实例，当workers > 1时，参数才生效 (default: {None})
        """

        if self._active:
            logger.warning("vxScheduler({self._name}) 已经启动，请勿重复启动...")
            return

        self._active = True

        channels = [vxChannel(channel_name, broker) for broker in self._bind_brokers]
        channels.append(vxChannel(channel_name))

        try:
            logger.info(f"vxScheduler({channel_name}) 开始运行...")

            if workers > 1:
                executor = executor or ThreadPoolExecutor(workers)
                with executor:
                    executor.map(self._run, channels)
            else:
                self._run(*channels)
        finally:
            self.stop()
            logger.info(f"vxScheduler({channel_name}) 停止运行...")

    def stop(self) -> None:
        """暂停函数"""
        self._active = False


def init_scheduler(
    scheduler_: vxSched,
    config_file: str = "./etc/config.json",
    mod_path: str = "mod/",
    default_config: Optional[Dict] = None,
) -> None:
    """vxsched 运行函数

    Keyword Arguments:
        scheduler {vxScheduler} -- scheduler 实例 (default: {'local'})
        config_file {str} -- 配置文件 (default: {'./etc/config.json'})
        mod_path {str} -- 模块目录 (default: {'mod/'})
        default_config {dict} -- 默认配置
    """

    context = scheduler_.context
    if default_config:
        context.update(**default_config)

    configfile = pathlib.Path(config_file)
    if not configfile.exists():
        logger.warning(f"配置文件不存在. {configfile.as_posix()}")
    else:
        context.load(configfile.as_posix())
        logger.info(f"加载配置文件{configfile.as_posix()}")

    logger.info("开始加载各类工具...")
    box = vxToolBox()
    tools_params = context.get("tools", None)
    if tools_params:
        logger.info("加载 tools:")
        box.loads(tools_params)
        logger.info("=" * 60)

    brokers_params = context.get("brokers", None)
    if brokers_params:
        logger.info("加载 brokers:")
        box.loads(brokers_params)
        for broker_id in brokers_params:
            scheduler_.bind_eventbroker(getattr(box, broker_id))
        logger.info("=" * 60)

    mod_dir = pathlib.Path(mod_path)
    if not mod_dir.exists():
        logger.warning(f"策略目录不存在:{mod_dir.as_posix()}")
    else:
        scheduler_.load_handlers(mod_dir.as_posix())

    return scheduler_


vxscheduler = vxSched()
