"""消息类型"""

from enum import Enum
from typing import Iterator, Optional, Any
from abc import ABCMeta, abstractmethod
from vxutils import to_timestring, to_json, vxtime, logger, vxJSONEncoder
from vxutils.dataclass import (
    vxDataClass,
    vxIntField,
    vxDatetimeField,
    vxField,
    vxUUIDField,
)


__all__ = ["vxEvent", "vxTrigger", "TriggerStatus"]


class TriggerStatus(Enum):
    """触发器状态

    Pending : 未开始
    Running : 已开始
    Completed: 已完成
    """

    Pending = 1  #  未开始
    Running = 2  #  已开始
    Completed = 3  #  已完成


class vxTrigger(Iterator[float], metaclass=ABCMeta):
    """触发器"""

    __slots__ = (
        "_last_trigger_dt",
        "_end_dt",
    )

    def __init__(self, end_dt=None):
        self._last_trigger_dt = None
        self._end_dt = end_dt or float("inf")

    @property
    def last_trigger_dt(self) -> Optional[float]:
        """最后一次执行日期，若未空，则未开始执行"""
        return self._last_trigger_dt

    @property
    def next_trigger_dt(self) -> Optional[float]:
        """下一次执行日期"""
        return self.get_next_trigger_dt()

    @property
    def end_dt(self) -> float:
        """下一次执行时间"""
        return self._end_dt

    @abstractmethod
    def get_next_trigger_dt(self) -> Optional[float]:
        """
        返回下一次执行timestamp，如果没有下次执行时间，则返回 'None'
        """

    def __getstate__(self) -> dict:
        """序列化函数"""
        return {key: getattr(self, key) for key in self.__slots__}

    def __setstate__(self, state: dict) -> None:
        """反序列化函数"""
        for key in self.__slots__:
            setattr(self, key, state[key])

    def __str__(self) -> str:
        return f"< {self.__class__.__name__} (id-{id(self)}): {to_json(self.message)}"

    __repr__ = __str__

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> float:
        now = vxtime.now()
        if self._end_dt and now > self._end_dt:
            raise StopIteration

        while self.get_next_trigger_dt() is not None:
            self._last_trigger_dt = self.get_next_trigger_dt()
            if self._last_trigger_dt > now:
                return self._last_trigger_dt

        raise StopIteration

    @property
    def message(self) -> dict:
        """trigger内部内容"""

        return {
            "last_trigger_dt": to_timestring(self.last_trigger_dt),
            "next_trigger_dt": to_timestring(self.get_next_trigger_dt()),
            "end_dt": to_timestring(self.end_dt),
            "status": self.status,
        }

    @property
    def status(self) -> TriggerStatus:
        """是否已完成"""

        if not self.last_trigger_dt:
            return TriggerStatus.Pending
        elif not self.next_trigger_dt:
            return TriggerStatus.Completed
        return TriggerStatus.Running


class vxEvent(vxDataClass):
    """消息类型"""

    __sortkeys__ = ("next_trigger_dt", "priority")

    # 消息id
    id: str = vxUUIDField(auto=True)
    # 消息通道
    channel: str = vxField("", str)
    # 消息类型
    type: str = vxField("", str)
    # 消息内容
    data: Any = vxField("")
    # 定时触发器
    trigger: Optional[vxTrigger] = vxField()
    # 下次触发事件
    next_trigger_dt: float = vxDatetimeField()
    # 优先级
    priority: int = vxIntField(10)


@vxJSONEncoder.register(vxTrigger)
def _(obj):
    return obj.message
